/* $Id: auswert.c,v 1.25 1998/12/28 16:08:57 mas Rel $
 * Copyright 1998 Marc Andr Selig.
 * All rights reserved.
 * Developed using showkey.c from kbd-0.96.
 * If compiled with USE_READLINE, requires the GNU readline library
 *	(tested with readline-2.2) and uses it for longer user input.
 * Programm zur Videoauswertung hnlich INTERACT fr Windows.
 * Voraussetzungen: timecoded mu laufen.
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Bugs:

 * Today (981206) I had a lot of trouble saving files.  Same problem
 * for Johanna 981227.  Don't really know what happens, but now use
 * perror() for error reporting.

 * Optically unique differentiation between "file write ok" and "file
   write error"

 * Should we really forget the file name on error?

 * Alt-F2 should ideally switch to vty2, but failing that, should at
   least not "save file" but print a message about how the user can
   switch vtys.

 * What happens if a file is being read in (with -f) that is larger than
   4000 lines?

 * PgUp sollte zunchst zum Anfang der Seite springen, wenn der Cursor
   nicht bereits dort ist; PgDn sollte ans Ende der Seite springen,
   wenn der Cursor nicht dort ist.

 * print_screen() only works correctly if each event takes up exactly
 * one line on the screen.  We should change it so that it will either
 * truncate lines to 80 columns or cope with longer lines.

 * When memory is full, we should really try to realloc() a larger
 * chunk of RAM instead of simply failing.  This would involve
 * dynamically allocating space for Rcolumn[] as well, which should
 * not be a problem, though.

 */

/* Ideas for future releases:
 
 * (Save more than one deleted event for re-insertion?  Real undo?)

 * Wanted: Search forward/backward in event description.
 * (Maybe F10 = search, Shift+F10 = search in specified column.)
 * (Can we search for regexps?  Do we want to do it at all [remember
 * our user base!]?  Simple substring matches might be the best
 * solution.)

 */



#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <stdlib.h>
#include <asm/io.h>
#include <assert.h>
#include <string.h>
#include "socklib.h"

#define DEFAULTSERVERHOST "localhost"

#ifdef USE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#define MAX_OUTPUT_LINES 4000	/* maximum number of events */

#define INPUT_BUFLEN 16

char pressed[128] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

char translations[] =  /* keycodes start at 1! */
{'?' /* 1:escape */,
 '1',
 '2',
 '3',
 '4',
 '5',
 '6',
 '7',
 '8',
 '9',
 '0',
 '',
 '\'',
 '?' /* 14:delete */,
 '?' /* 15:tab */,
 'q',
 'w',
 'e',
 'r',
 't',
 'z',
 'u',
 'i',
 'o',
 'p',
 '',
 '+',
 '?' /* 28:return */,
 '?' /* 29:control */,
 'a',
 's',
 'd',
 'f',
 'g',
 'h',
 'j',
 'k',
 'l',
 '',
 '',
 '^',
 '?' /* 42:shift */,
 '#',
 'y',
 'x',
 'c',
 'v',
 'b',
 'n',
 'm',
 ',',
 '.',
 '-',
 '?' /* 54:shift */,
 '*' /* kp */,
 '?' /* 56:alt */,
 ' ',
 '?' /* 58:capslock */,
 '?', '?', '?', '?', '?', '?', '?', '?', '?', '?' /* 59-68:F1-F10 */,
 '?' /* 69:num lock */,
 '?' /* 70: scroll lock */,
 '7' /* kp */,
 '8' /* kp */,
 '9' /* kp */,
 '-' /* kp */,
 '4' /* kp */,
 '5' /* kp */,
 '6' /* kp */,
 '+' /* kp */,
 '1' /* kp */,
 '2' /* kp */,
 '3' /* kp */,
 '0' /* kp */,

 ',' /* kp */,
 '?' /* 84: last_console */,
 '?' /* 85: undefined */,
 '<',
 '?' /* 87: F11 */,
 '?' /* 88: F12 */,
 '?', '?', '?', '?', '?', '?', '?' /* 89-95: undef */,
 '?' /* 96: kp_enter */,
 '?' /* 97: control */,
 '/' /* kp */,
 '\\' /* 99: control_backslash */,
 '?' /* 100: altgr */,
 '?' /* 101: break */,
 '?' /* 102: find = home */,
 '?' /* 103: up */,
 '?' /* 104: prior = PageUp */,
 '?' /* 105: left */,
 '?' /* 106: right */,
 '?' /* 107: select = end */,
 '?' /* 108: down */,
 '?' /* 109: next = PageDown */,
 '?' /* 110: insert */,
 '?' /* 111: remove */,
 '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?', '?',
 '?', '?', '?' /* 112-128 do not exist, but are possible in principle */
};

char timecodes[16*128];		/* timecodes at key down are recorded here */
	/* There's room for exactly one timecode per key.  Timecodes
	 * are aligned at 16 byte boundaries.  13 bytes would in
	 * theory be quite enough. */

char outfilename[256];		/* filename for saving */

int tmp;	/* for debugging of tcsetattr */

extern int getfd();
int fd, sock;
int oldkbmode;
struct termios old;

unsigned char *Rentry, *Rexit, *Rtext;
	/* Completed events ("R"esults) are stored in here. */
	/* An event is flagged as "deleted" by setting the first byte
	 * of Rentry for that event to zero! */
int Rcolumn [MAX_OUTPUT_LINES];
	/* Says which column Rtext for the corresponding event should be in.
	 * If Rcolumn[] == 1, Rtext[] follows with one tab distance after
	 * the timecodes, so you might say it is in column 3.  If
	 * Rcolumn[] == 3, there are three tabs total between the exit
	 * timecode and Rtext[]. */

#define TC_LEN 12	/* length of the text for a time code + \0 */
#define TEXT_LEN 80	/* length of the descriptional text */

int cursor, screen_start, last_result, screen_lines, have_deletions,
  default_coding_column;



void tcp_get_timecode (char *target, int buflen)
{
  char buf[512];

  write_to_socket(sock, "GETTC\n");
  if (read_from_socket(sock, buf, sizeof(buf)) < 0)
    strncpy (target, "tcp error !", buflen);
  else {
    strncpy (target, buf, buflen);
    if (strchr (target, '\n'))
      *strchr (target, '\n') = 0;
  }
}



void restore_kbmode(void) {
  /* restore the keyboard mode so we can type again */
  if (ioctl(fd, KDSKBMODE, oldkbmode)) {
    perror("KDSKBMODE");
    exit(1);
  }
  tmp = tcsetattr(fd, 0, &old);
  if (tmp)
    printf("tcsetattr = %d\n", tmp);
  close(fd);
}



void set_kbmode(void) {
  struct termios new;

  fd = getfd();

  /* Save previous keyboard mode, which may not be XLATE if we run under X. */
  if (ioctl(fd, KDGKBMODE, &oldkbmode)) {
    perror("KDGKBMODE");
    exit(1);
  }
  if (oldkbmode != K_XLATE) {
    printf("[ If you are trying this under X, it might not work\n");
    printf("since the X server is also reading /dev/console. ]\n");
  }

  /* Preserve previous keyboard mode */
  tmp = tcgetattr(fd, &old);
  if (tmp)
    printf("tcgetattr = %d\n", tmp);
  tmp = tcgetattr(fd, &new);
  if (tmp)
    printf("tcgetattr = %d\n", tmp);

  /* Set keyboard mode to medium raw so we can read individual keycodes,
   * can handle several keys pressed at the same time and don't have to
   * wait for the key(s) to be released before we can see it/them. */
  new.c_lflag &= ~ (ICANON | ECHO | ISIG);
  new.c_iflag = 0;
  new.c_cc[VMIN] = INPUT_BUFLEN;
  new.c_cc[VTIME] = 1;	/* 0.1 sec intercharacter timeout */
  tmp = tcsetattr(fd, TCSAFLUSH, &new);
  if (tmp)
    printf("tcsetattr = %d\n", tmp);
  if (ioctl(fd, KDSKBMODE, K_MEDIUMRAW)) {
    perror("KDSKBMODE");
    exit(1);
  }
}



char *get_some_tabs (int how_many) {
  /* Returns a pointer to how_many tabs. */
  static char heap_of_tabs[] = "\t\t\t\t\t\t\t\t\t\t";

  if ((how_many < 1) || (how_many > sizeof(heap_of_tabs)))
    return NULL;

  return heap_of_tabs + sizeof(heap_of_tabs) - how_many - 1;
}



void get_file_name() {
  /* Dateinamen fr die Zieldatei einlesen */

#ifdef USE_READLINE
  char *inbuffer;
#endif

  restore_kbmode();
  printf("Enter file name: ");

#ifdef USE_READLINE

  inbuffer = readline(NULL);
  /* readline() does erase a trailing newline! */

  strncpy(outfilename, inbuffer, sizeof(outfilename) - 1);
  *(outfilename + sizeof(outfilename) - 1) = 0;

  if (inbuffer)
    free(inbuffer);	/* was allocated by readline() */

#else

  fgets(outfilename, sizeof(outfilename), stdin);
  if (strchr(outfilename, '\n'))
    *strchr(outfilename, '\n') = 0;

#endif

  set_kbmode();
}



int save_file() {
  FILE *outfile;
  int i;
  char backupfile[sizeof(outfilename)+1];

  printf("\n");

  while (! (*outfilename))	/* Read filename if we don't know it yet */
    get_file_name();

  /* Now create a backup of that file if it already exists. */

  strcpy (backupfile, outfilename); /* Won't need strncpy or strncat here */
  strcat (backupfile, "~");	/* since backupfile was initialized to just */
  /* the right length and outfilename contains a 0 at the end (man fgets!). */

  if (!rename (outfilename, backupfile))
    printf("Keeping a backup of %s as %s.\n", outfilename, backupfile);

  /* Now save it. */

  if ((outfile = fopen(outfilename, "w")) == NULL) {
    perror("fopen (file could not be saved)");
    *outfilename=0;
    printf("(I will now forget the output file name; you will have to enter it again!)\n");
    printf("(Messages will be cleared once you resume your work...)\n\n");
    return 1;
  }

  printf("Writing to %s... ", outfilename);

  for (i=0; i<last_result; i++) {
    if (*(Rentry + i*TC_LEN)) {
      if (! fprintf(outfile, "%s\t%s%s%s\n",
		    Rentry + i*TC_LEN,
		    Rexit + i*TC_LEN,
		    get_some_tabs(Rcolumn[i]),
		    Rtext + i*TEXT_LEN)) {
	printf("fprintf failed for event %d, file could not be saved\n", i);
	fclose(outfile);
	*outfilename=0;
	printf("(I will now forget the output file name; you will have to enter it again!)\n");
	printf("(Messages will be cleared once you resume your work...)\n\n");
	return 1;
      }
    }
  }

  if (fclose(outfile)) {
    printf("fclose failed, file could not be saved\n");
    *outfilename=0;
    printf("(I will now forget the output file name; you will have to enter it again!)\n");
    printf("(Messages will be cleared once you resume your work...)\n\n");
    return 1;
  }

  printf("OK\n");
  printf("(Messages will be cleared once you resume your work...)\n\n");

  return 0;
}



int read_file (char * infilename) {
  /* returns 1 on error */

  FILE *infile;
  char inbuffer[256];
  char *exitpointer, *textpointer;

  if ((infile = fopen(infilename, "r")) == NULL) {
    perror("fopen (file could not be opened)");
    return 1;
  }

  printf("Reading file %s... ", infilename);

  last_result = 0;

  /* Now we
   * - replace the first two tabs in inbuffer by \0,
   * - save all text before these two as entry and exit time code repectively,
   * - count the number of tabs immediately after the second tab
   * - and save any text _after_ those tabs as the event description.
   * If any fields are missing, we try to handle this courteously. */

  while(fgets(inbuffer, sizeof(inbuffer), infile) != NULL) {
    Rcolumn[last_result] = 1;
    /* There must be at least one tab between the time codes and the text. */

    if ((exitpointer = strchr(inbuffer, '\t')) != NULL) {
      *(exitpointer++) = 0;	/* replace \t by \0 */
      strncpy(Rentry + last_result*TC_LEN, inbuffer, TC_LEN-1);
      *(Rentry + (last_result+1)*TC_LEN - 1) = 0;
        /* strncpy does not null-terminate in all cases! */

      if ((textpointer = strchr(exitpointer, '\t')) != NULL) {
	*(textpointer++) = 0;	/* replace \t by \0 */
	strncpy(Rexit + last_result*TC_LEN, exitpointer, TC_LEN-1);
	*(Rexit + (last_result+1)*TC_LEN - 1) = 0;

	while (*textpointer == '\t') {	/* count tabs between tc's and text */
	  Rcolumn[last_result]++;
	  textpointer++;
	}

	if (strchr(textpointer, '\n') != NULL)	/* remove trailing \n */
	  *(strchr(textpointer, '\n')) = 0;
	strncpy(Rtext + last_result*TEXT_LEN, textpointer, TEXT_LEN-1);
	*(Rtext + (last_result+1)*TEXT_LEN - 1) = 0;

      } else {	/* only two tokens per line */
	strncpy(Rexit + last_result*TC_LEN, exitpointer, TC_LEN-1);
	*(Rexit + (last_result+1)*TC_LEN -1 ) = 0;
	*(Rtext + last_result*TEXT_LEN) = 0;
      }

    } else {	/* only one token per line */
      strncpy(Rentry + last_result*TC_LEN, inbuffer, TC_LEN-1);
      *(Rentry + (last_result+1)*TC_LEN - 1) = 0;
      *(Rexit + last_result*TC_LEN) = 0;
      *(Rtext + last_result*TEXT_LEN) = 0;
    }
    last_result++;
  }

  fclose(infile);

  printf(" OK (saving this file name for output)\n");

  strcpy(outfilename, infilename);

  return 0;
}



void clean_up(void) {
  restore_kbmode();
  free(Rentry);
  free(Rexit);
  free(Rtext);
  close(sock);
  sync();
}



void die(int x) {
  printf("Caught signal %d, cleaning up...\n", x);
  if (x == SIGPIPE)
    printf("This probably means that the connection to our time code server was lost.\n");
  printf("I will try to save your work, but since I am dying, I might not succeed.\n");
  save_file();
  clean_up();
  exit(1);
}



int screen_lines_between (int start, int finish) {
  /* Calculates number of visible lines needed to display start to finish */
  int i, lines_seen;

  lines_seen = 0;

  for (i=start; i<=finish; i++)
    if (*(Rentry + i*TC_LEN))
      lines_seen++;

  return lines_seen;
}



void print_screen () {
  /* prints a screenful of recorded events and the cursor */
  int i, lines_printed;

  printf("\n%s%d event%s; default coding column %d\n",
	 (last_result >= MAX_OUTPUT_LINES) ? " \aMEMORY IS FULL!  --  " : "",
	 last_result-have_deletions,
	 (last_result-have_deletions == 1) ? "" : "s",	/* "events"/"event"? */
	 default_coding_column+2);

  i = screen_start;
  lines_printed = 0;
  while ((lines_printed < screen_lines) && (i < last_result)) {
    if (*(Rentry + i*TC_LEN)) {
      printf ("%s%s\t%s%s%s\n",
	      Rentry + i*TC_LEN,
	      (i == cursor) ? " -->" : "",
	      Rexit + i*TC_LEN,
	      get_some_tabs(Rcolumn[i]),
	      Rtext + i*TEXT_LEN);
      lines_printed++;
    }
    i++;
  }
  for (i = lines_printed; i < screen_lines; i++)
    printf("\n");
}



void move_cursor_up(void) {
  /* move cursor one line up, allowing for deleted lines */

  do	      /* we don't want to arrive at a deleted event */
    cursor--;
  while ((cursor >= 0) && (*(Rentry+cursor*TC_LEN) == 0));
  if (cursor < 0)
    cursor = 0;
  if (cursor < screen_start)
    screen_start = cursor;
}



void adjust_cursor_up(void) {
  /* move cursor one line up if current event has been deleted */

  if (*(Rentry + cursor*TC_LEN) == 0)
    move_cursor_up();
}



void move_cursor_down(void) {
  /* move cursor one line down, allowing for deleted lines */

  do	      /* we don't want to arrive at a deleted event */
    cursor++;
  while ((cursor < last_result) && (*(Rentry+cursor*TC_LEN) == 0));
  if (cursor >= last_result)
    cursor = last_result - 1;
  while (screen_lines_between(screen_start, cursor) > screen_lines)
    screen_start++;
}



void adjust_cursor_down(void) {
  /* move cursor one line down if current event has been deleted */

  if (*(Rentry + cursor*TC_LEN) == 0)
    move_cursor_down();
}



int enter_text (char *target, int max_len, int remove_tabs) {
  /* Enter text, e.g. a time code or the textual description.  This
   * function takes care of the keyboard mode, but not of a prompt.
   * If the user enters no text (just presses return), target is
   * unchanged; in this case, the function returns 0. */

#ifdef USE_READLINE
  char *inbuffer;
#else
  char inbuffer[256];
#endif
  int return_code = 1;

  restore_kbmode();	/* get out of medium raw mode */

#ifdef USE_READLINE
  inbuffer = readline(NULL);
  /* readline() does erase a trailing newline! */
#else
  fgets(inbuffer, 256, stdin);
  /* We read into a buffer in case the user inputs more characters
   * than we need.  This way, we can simply discard the rest; if we
   * read directly into target, the surplus characters would be left
   * on the input stream. */

  /* fgets does not erase a trailing newline, so we do it */
  if (strchr(inbuffer, '\n'))
    *strchr(inbuffer, '\n') = 0;
#endif

  if ((inbuffer) && (strlen(inbuffer) > 0)) {
    strncpy(target, inbuffer, max_len - 1);
    *(target + max_len - 1) = 0;
    /* strncpy does not null-terminate in all cases! */

    if (remove_tabs) {
      /* Replace all tabs by spaces */
      while (strchr(target, '\t'))
	*strchr(target, '\t') = ' ';
    }
  } else
    return_code = 0;

#ifdef USE_READLINE
  if (inbuffer)
    free(inbuffer);	/* was allocated by readline() */
#endif

  set_kbmode();	/* return keyboard to medium raw */

  return return_code;
}



#define special_key(a) (((a) == 42) || ((a) == 54))



int main (int argc, char *argv[]) {
  unsigned char buf[INPUT_BUFLEN];
  unsigned char tc[16];
  int i, n, quit_now, print_now;
  char *getenv_lines;
  char last_text[TEXT_LEN];	/* for remembering user input at `Enter' */
  int last_column;
  struct sockaddr_in servername;
  char user_servername [512];

  char remember_entry[TC_LEN], remember_exit[TC_LEN],
    remember_text[TEXT_LEN], remember_column;
	/* deleted events get stored in here */

  assert ((Rentry = calloc(MAX_OUTPUT_LINES, TC_LEN)) != NULL);
  assert ((Rexit  = calloc(MAX_OUTPUT_LINES, TC_LEN)) != NULL);
  assert ((Rtext  = calloc(MAX_OUTPUT_LINES, TEXT_LEN)) != NULL);
  /* calloc initializes all blocks to zero! */

  quit_now = 0;
  print_now = 1;
  cursor = 0;
  screen_start = 0;
  last_result = 0;
  have_deletions = 0;
  outfilename[0] = 0;
  default_coding_column = 1;	/* at least 1 tab between time code and text */
  *last_text = 0;
  last_column = 0;
  *remember_entry = 0;
  *remember_exit = 0;
  *remember_text = 0;
  remember_column = 0;
  strncpy (user_servername, DEFAULTSERVERHOST, sizeof(user_servername));

  getenv_lines = getenv("LINES");
  if ((!getenv_lines) || (!sscanf (getenv_lines, "%d", &screen_lines)))
    screen_lines = 25;		/* default number of screen lines */
  screen_lines-=2;	/* one header above and one empty line below data */

  while ((n = getopt(argc, argv, "f:s:h?")) != -1)
    switch (n) {
    case 'f':
      if (read_file(optarg))
	exit(1);
      break;

    case 's':
      strncpy (user_servername, optarg, sizeof(user_servername));
      printf("Will connect to server on %s.\n", user_servername);
      break;

    case 'h':
    case '?':
      printf("\n$Id: auswert.c,v 1.25 1998/12/28 16:08:57 mas Rel $\n\n");

      printf("Copyright 1998 Marc Andr Selig.\n");
      printf("auswert comes with ABSOLUTELY NO WARRANTY; for details see file COPYING.\n");
      printf("This is free software, and you are welcome to redistribute it under\n");
      printf("certain conditions; see file COPYING for details.\n\n");

      printf("Usage: auswert [-f infile] [-n] [-h|-?]\n");
      printf("  -f infile\n");
      printf("\tread infile to initialize data structures\n");
      printf("  -s serverhost\n");
      printf("\tconnect to server on serverhost (default: %s)\n",
	     DEFAULTSERVERHOST);
      printf("  -h|-?\n");
      printf("\tprint this help screen\n\n");
      exit(0);
    }

  /*
    if we receive a signal, we want to exit nicely, in
    order not to leave the keyboard in an unusable mode
    */
  signal(SIGHUP, die);
  signal(SIGINT, die);
  signal(SIGQUIT, die);
  signal(SIGILL, die);
  signal(SIGTRAP, die);
  signal(SIGABRT, die);
  signal(SIGIOT, die);
  signal(SIGFPE, die);
  signal(SIGKILL, die);
  signal(SIGUSR1, die);
  signal(SIGSEGV, die);
  signal(SIGUSR2, die);
  signal(SIGPIPE, die);
  signal(SIGTERM, die);
#ifdef SIGSTKFLT
  signal(SIGSTKFLT, die);
#endif
  signal(SIGCHLD, die);
  signal(SIGCONT, die);
  signal(SIGSTOP, die);
  signal(SIGTSTP, die);
  signal(SIGTTIN, die);
  signal(SIGTTOU, die);

  /* Create the socket. */
  sock = socket (PF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
    perror ("socket (auswert)");
    exit (EXIT_FAILURE);
  }

  /* Connect to the server. */
  init_sockaddr (&servername, user_servername, INET_PORT);
  if (0 > connect (sock,
		   (struct sockaddr *) &servername,
		   sizeof (servername))) {
    perror ("connect (client)");
    exit (EXIT_FAILURE);
  }

  /* Set keyboard to medium raw mode */
  set_kbmode();

  /* main program loop */
  while (!quit_now) {
    if (print_now) {
      print_screen();
      print_now = 0;
    }
    n = read(fd, buf, sizeof(buf));
    for (i = 0; i < n; i++) {
      if (buf[i] & 0x80) {	/* key released */
	if ((pressed[buf[i] & 0x7f]) &&	/* mask for keycode */
	    (!(special_key(buf[i] & 0x7f)))) {	/* no shift etc.! */
	  if (last_result < MAX_OUTPUT_LINES) {
	    tcp_get_timecode(tc, sizeof(tc));

	    /* Remember all relevant data for the new event. */
	    strncpy(Rentry + last_result*TC_LEN,
		    &timecodes[(buf[i] & 0x7f)*16],
		    TC_LEN);
	    strncpy(Rexit + last_result*TC_LEN, tc, TC_LEN);
	    Rcolumn[last_result] = default_coding_column;
	    *(Rtext + last_result*TEXT_LEN) =
	      translations[(buf[i] & 0x7f)-1];
	    *(Rtext + last_result*TEXT_LEN + 1) = 0;

	    cursor = last_result;		/* cursor to new event */
	    while (screen_lines_between (screen_start, cursor) > screen_lines)
	      screen_start++;
	    last_result++;
	  } else
	    printf("\a");
	  print_now = 1;
	  /* The screen is not printed after every single key released.
	   * This saves some time and makes it possible to handle users'
	   * quickstep dances on the keyboard adequately. */
	}
	pressed[buf[i] & 0x7f] = 0;	/* clear "key is pressed" flag */
      } else {		/* key pressed */
	if (!pressed[buf[i] & 0x7f]) {
	  /* Note that special keys are NOT normally flagged as
	   * pressed (exception: shift), so you can press a cursor key
	   * and hold it and get it to repeat.  Normal keys, on the
	   * other hand, do not repeat: holding them just means a
	   * longer event. */
	  switch (buf[i] & 0x7f) {	/* handle some keys specially */

	  case 28:			/* `return': change event text */
			/* or, with shift, set last-entered event description
			 * (incl. column) for current event */
	    if (*(Rentry+cursor*TC_LEN)) {
	      if (pressed[42] || pressed[54]) {
		/* with shift: set current event text+column to the
		 * one last entered */
		if (last_column > 0) {
		  strcpy (Rtext+cursor*TEXT_LEN, last_text);
		  Rcolumn[cursor] = last_column;
		  print_now = 1;
		} else {
		  printf("No text saved yet!\n");
		  printf("(Just continue your work to clear this message.)\n");
		}
	      } else {
		/* without shift: change text for current event */
		printf("Enter new text for event [%s]: ",
		       Rtext + cursor*TEXT_LEN);
		if (enter_text (Rtext+cursor*TEXT_LEN, TEXT_LEN, 0)) {
		  /* remember what was entered */
		  strcpy (last_text, Rtext+cursor*TEXT_LEN);
		  last_column = Rcolumn[cursor];
		}
		print_now = 1;
	      }
	    }
	    break;

	  case 42:			/* Left  Shift */
	    pressed[42] = 1;	/* mark as pressed */
	    /* We do it this way instead of marking both shift keys as
	     * pressed or having a separate flag so that clearing the
	     * keys does not require special code. */
	    break;

	  case 54:			/* Right Shift */
	    pressed[54] = 1;
	    break;

	  case 59:			/* F1: help */
	    printf("\nPrinting keys generate an event, cursor keys move vertically.  Others:\n");
	    printf("\nKey\tFunction\n---\t--------\n");
	    printf("F1\tHelp\n");
	    printf("F2\tSave (with shift: save and exit)\n");
	    printf("F3\tMove event description to column 3 (with shift: make col 3 default)\n");
	    printf("F4\tMove event description to column 4 (with shift: make col 4 default)\n");
	    printf("F5\tMove event description to column 5 (with shift: make col 5 default)\n");
	    printf("F6\tMove event description to column 6 (with shift: make col 6 default)\n");
	    printf("F7\tMove event description to column 7 (with shift: make col 7 default)\n");
	    printf("F8\tMove event description to column 8 (with shift: make col 8 default)\n");
	    printf("F11\tChange entry time code for event at cursor position\n");
	    printf("\t(with shift: set it to the current time code)\n");
	    printf("F12\tChange exit time code  for event at cursor position\n");
	    printf("\t(with shift: set it to the current time code)\n");
	    printf("Return\tChange description     for event at cursor position\n");
	    printf("\t(with shift: set it to last-entered event description, incl. column)\n");
	    printf("DEL\tRemove                     event at cursor position\n");
	    printf("\nResuming your work will clear this help notice.\n\n");
	    break;

	  case 60:			/* F2: save */
	    if (!save_file())
	      if (pressed[42] || pressed[54])	/* with shift: save and exit */
		quit_now = 1;
	    break;

	  case 61:	/* F3: move text to column 3 or make column 3
			 * the default coding column */
	  case 62:	/* same for F4 .. F8 */
	  case 63:
	  case 64:
	  case 65:
	  case 66:
	    if (pressed[42] || pressed[54])
	      default_coding_column = (buf[i] & 0x7f) - 60;
	    else {
	      Rcolumn[cursor] = (buf[i] & 0x7f) - 60;
	    }
	    print_now = 1;
	    break;

	  case 87:			/* F11: change entry time code */
	    if (*(Rentry+cursor*TC_LEN)) {
	      if (pressed[42] || pressed[54])	/* shift: set tc to current */
		tcp_get_timecode(Rentry+cursor*TC_LEN, TC_LEN);
	      else {
		printf("Enter new entry time code [%s]: ",
		       Rentry + cursor*TC_LEN);
		enter_text (Rentry+cursor*TC_LEN, TC_LEN, 1);
	      }
	      print_now = 1;
	    }
	    break;

	  case 88:			/* F12: change exit time code */
	    if (*(Rentry+cursor*TC_LEN)) {
	      if (pressed[42] || pressed[54])	/* shift: set tc to current */
		tcp_get_timecode(Rexit+cursor*TC_LEN, TC_LEN);
	      else {
		printf("Enter new exit time code [%s]: ",
		       Rexit + cursor*TC_LEN);
		enter_text (Rexit+cursor*TC_LEN, TC_LEN, 1);
	      }
	      print_now = 1;
	    }
	    break;

	  case 102:			/* find = home */
	    cursor = 0;
	    screen_start = 0;
	    adjust_cursor_down();
	    print_now = 1;
	    break;

	  case 103:			/* move cursor up */
	    move_cursor_up();
	    adjust_cursor_down();
	    print_now = 1;
	    break;

	  case 104:			/* page up */
	    for (i=0; i < screen_lines - 1; i++)
	      move_cursor_up();
	    adjust_cursor_down();
	    print_now = 1;
	    break;

	  case 107:			/* select = end */
	    cursor = last_result;
	    screen_start = last_result;
	    adjust_cursor_up();
	    while ((screen_lines_between (screen_start, cursor) < screen_lines)
		   && (screen_start > 0))
	      screen_start--;
	    print_now = 1;
	    break;

	  case 108:			/* move cursor down */
	    move_cursor_down();
	    adjust_cursor_up();
	    print_now = 1;
	    break;

	  case 109:			/* page down */
	    for (i=0; i < screen_lines - 1; i++)
	      move_cursor_down();
	    adjust_cursor_up();
	    print_now = 1;
	    break;

	  case 110:			/* Ins: insert last deleted event */
	    if (*remember_entry != 0) {
	      strcpy(Rentry+last_result*TC_LEN,   remember_entry);
	      strcpy(Rexit +last_result*TC_LEN,   remember_exit);
	      strcpy(Rtext +last_result*TEXT_LEN, remember_text);
	      Rcolumn[last_result] = remember_column;

	      /* Mark remembered event as empty; we don't need it any more. */
	      *remember_entry = 0;

	      cursor = last_result;		/* cursor to new event */
	      while (screen_lines_between(screen_start, cursor) > screen_lines)
		screen_start++;
	      last_result++;

	      print_now = 1;
	    }
	    break;

	  case 111:			/* Del: remove current element */
	    if (*(Rentry + cursor*TC_LEN) != 0) {
	      /* Remember event */
	      strcpy(remember_entry, Rentry+cursor*TC_LEN);
	      strcpy(remember_exit,  Rexit +cursor*TC_LEN);
	      strcpy(remember_text,  Rtext +cursor*TEXT_LEN);
	      remember_column = Rcolumn[cursor];

	      /* This will mark the current event as deleted */
	      *(Rentry+cursor*TC_LEN) = 0;
	      have_deletions++;
	      move_cursor_down();	/* move away from the empty element */
	      adjust_cursor_up();
	      print_now = 1;
	    }
	    break;

	  default:		/* else it's the beginning of an event */
	    if (translations[(buf[i] & 0x7f)-1] != '?') {
	      /* special keys are flagged as '?'; don't accept them */
	      pressed[buf[i] & 0x7f] = 1;	/* flag key as "pressed" */
	      tcp_get_timecode(&timecodes[(buf[i] & 0x7f)*16], 16); /* remember tc */
	    }
	  }
	}
      }
    }
  }

  clean_up();
  exit(0);
}
